#! /usr/bin/env python
#
#   gtopology.py -- Front-end script for running topology evaluations in
#   python over different initial parameter conditions.
#
#   Copyright (C) 2016, 2019  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple topology jobs written in python.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gtopology.py --help`` for program usage
instructions.

Input parameters consists of:
10,10,2,1,10
10,10,3,1,10
10,10,4,1,10
...

XXX: To be clarified:
* dependency in igraph python library
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-05-10:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gtopology
    gtopology.GtopologyScript().run()

import os
import sys
import time
import tempfile

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_MASTER_SCRIPT="run_topologies_cluster.py"

## custom application class
class GtopologyApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gtopology'

    def __init__(self, param_file, source_folder, **extra_args):

        # setup output
        outputs = dict()

        # setup input
        inputs = dict()
        inputs[param_file] = "./input.csv"
        inputs[source_folder] = "./source/"

        if 'master_script' in extra_args:
            master_script = extra_args['master_script']
        else:
            master_script = DEFAULT_MASTER_SCRIPT

        arguments ="pwd ; date ; cd source ; python %s ../input.csv " % master_script

        # prepare execution script from command
        execution_script = """
#!/bin/sh

export PYTHONPATH=./source:$PYTHONPATH
cd source
# execute command
python %s ../input.csv
RET=$?

echo Program terminated with exit code $RET
exit $RET
        """ % (master_script)

        try:
            # create script file
            (handle, self.tmp_filename) = tempfile.mkstemp(prefix='gtopology-', suffix=extra_args['jobname'])

            # XXX: use NamedTemporaryFile instead with 'delete' = False

            fd = open(self.tmp_filename,'w')
            fd.write(execution_script)
            fd.close()
            os.chmod(fd.name,0o777)
        except Exception, ex:
            gc3libs.log.debug("Error creating execution script" +
                              "Error type: %s." % type(ex) +
                              "Message: %s"  %ex.message)
            raise

        inputs[fd.name] = "./master.sh"

        Application.__init__(
            self,
            arguments = ['./master.sh'],
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gtopology.log',
            join=True,
            executables = ['./master.sh'],
            **extra_args)

class GtopologyScript(SessionBasedScript):
    """
    Splits input .csv file into smaller chunks, each of them of size
    'self.params.chunk_size'.
    Then it submits one execution for each of the created chunked files.

    The ``gtopology`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gtopology`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gtopology``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GtopologyApplication,
            stats_only_for = GtopologyApplication,
            )

    def setup_options(self):
        self.add_param("-k", "--chunk", metavar="[NUM]",
                       dest="chunk_size", default="1000",
                       help="How to split the edges input data set. "
                       "Default: %(default)s")

        self.add_param("-M", "--master", metavar="[PATH]",
                       dest="master_script", default=DEFAULT_MASTER_SCRIPT,
                       help="Name of the main execution script. "
                       "Default: %(default)s")


    def setup_args(self):

        self.add_param('input', type=str,
                       help="Input csv.")

        self.add_param('source', type=str,
                       help="Location of the main script source folder.")


    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        try:
            assert os.path.isfile(self.params.input), \
                "Input csv file '%s' not found" % self.params.input

            assert os.path.isdir(self.params.source), \
                "Input source folder '%s' not found" % self.params.source

        except AssertionError as ex:
            raise OSError(ex.message)

        # Verify that 'self.params.chunk_size' is int
        int(self.params.chunk_size)
        self.params.chunk_size = int(self.params.chunk_size)


    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """
        tasks = []

        for (input_file, index_chunk) in self._generate_chunked_files_and_list(self.params.input,
                                                                              self.params.chunk_size):
            extra_args = extra.copy()

            jobname = "gtopology-%s" % (str(index_chunk))
            extra_args['jobname'] = jobname
            extra_args['index_chunk'] = str(index_chunk)

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        os.path.join('.computation',
                                                                                     jobname))

            extra_args['source'] = self.params.source
            extra_args['master_script'] = self.params.master_script

            self.log.debug("Creating Task for index : %d - %d" %
                           (index_chunk, (index_chunk + self.params.chunk_size)))

            tasks.append(GtopologyApplication(
                input_file,
                os.path.abspath(self.params.source),
                **extra_args))

        return tasks

    def _generate_chunked_files_and_list(self, file_to_chunk, chunk_size=1000):
        """
        Takes a file_name as input and a defined chunk size
        ( uses 1000 as default )
        returns a list of filenames 1 for each chunk created and a corresponding
        index reference
        e.g. ('/tmp/chunk2800.csv,2800) for the chunk segment that goes
        from 2800 to (2800 + chunk_size)
        """
        index = 0
        chunk = []
        failure = False
        chunk_files_dir = os.path.join(self.session.path,"tmp")

        # creating 'chunk_files_dir'
        if not(os.path.isdir(chunk_files_dir)):
            try:
                os.mkdir(chunk_files_dir)
            except OSError, osx:
                gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                                  "Error %s." % str(osx) +
                                  "Using default '/tmp'")
                chunk_files_dir = "/tmp"

        try:
            fd = open(file_to_chunk,'rb')

            fout = None

            for (i, line) in enumerate(fd):
                if i % chunk_size == 0:
                    if fout:
                        fout.close()
                    (handle, self.tmp_filename) = tempfile.mkstemp(dir=chunk_files_dir,
                                                                    prefix=
                                                                   'gtopology-',
                                                                   suffix=
                                                                   "%d.csv" % i)
                    fout = open(self.tmp_filename,'w')
                    chunk.append((fout.name,i))
                fout.write(line)
            fout.close()
        except OSError, osx:
            gc3libs.log.critical("Failed while creating chunk files." +
                                 "Error %s", (str(osx)))
            failure = True
        finally:
            if failure:
                # remove all tmp file created
                gc3libs.log.info("Could not generate full chunk list. "
                                 "Removing all existing tmp files... ")
                for (cfile, index) in chunk:
                    try:
                        os.remove(cfile)
                    except OSError, osx:
                        gc3libs.log.error("Failed while removing " +
                                          "tmp file %s. " +
                                          "Message %s" % osx.message)

        return chunk



#!/usr/bin/env python

"""
export SUBJECTS_DIR=/our/out/data/path

************
SUBJECT: s01
************
trac-all -prep -c {dmrirc} -debug
trac-all -bedp -c {dmrirc}
trac-all -path -c {dmrirc}
"""

from __future__ import absolute_import, print_function, unicode_literals
import sys
import os
import subprocess

FSAVERAGE = "fsaverage"
FS_SUBJECT_FSAVERAGE = os.path.join(os.environ.get("FREESURFER_HOME", ''),"subjects",FSAVERAGE)

TRAC_CMD_STEP1="trac-all -prep -c {dmrirc} -debug"
TRAC_CMD_STEP2="trac-all -bedp -c {dmrirc}"
TRAC_CMD_STEP3="trac-all -path -c {dmrirc}"

TRAC_PIPELINE = [ TRAC_CMD_STEP1,
                  TRAC_CMD_STEP2,
                  TRAC_CMD_STEP3 ]

def Usage():
    print ("Usage: gtrac_wrapper.py <dmrirc file>")

def RunTrac(dmrirc_input):
    """
    For a given subject dir, execute the 3 trac steps sequentially.
    """

    # Verify input arguments
    assert os.path.isfile(dmrirc_input)

    ret = 0

    for step in TRAC_PIPELINE:
        cmd = step.format(dmrirc=dmrirc_input)
        print("Running '%s' " % cmd)
        (ret,stdout,stderr) = runme(cmd)
        if ret != 0:
            print("[failed]")
            print("Execution failed with exit code: %d" % ret)
            print("Output message: %s" % stdout)
            print("Error message: %s" % stderr)
            break
        else:
            print "[ok]"
    return ret

def runme(command):
    """
    Comodity function to run commands using `subprocess` module
    Input: command to run
    Output: none
    Raise Exception in case command fails
    """
    proc = subprocess.Popen(
        [command],
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE)

    (stdout, stderr) = proc.communicate()
    return (proc.returncode, stdout, stderr)

if __name__ == '__main__':
    if (len(sys.argv) != 2):
        sys.exit(Usage())
    sys.exit(RunTrac(sys.argv[1]))



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2012, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

# System imports
import os
import getpass
from tempfile import NamedTemporaryFile

# Nose imports
import pytest

# GC3 imports
from gc3libs.backends import transport
from gc3libs.exceptions import TransportError


class StubForTestTransport(object):

    def extra_setup(self):
        (exitcode, tmpdir, stderror) = self.transport.execute_command(
            'mktemp -d /tmp/test_transport.XXXXXXXXX')
        self.tmpdir = tmpdir.strip()

    def tearDown(self):
        self.transport.remove_tree(self.tmpdir)
        self.transport.close()

    def test_get_remote_username(self):
        user = getpass.getuser()
        assert user == self.transport.get_remote_username()

    def test_isdir(self):
        assert self.transport.isdir(self.tmpdir)

    def test_listdir(self):
        assert len(self.transport.listdir(self.tmpdir)) == 0

    def test_makedirs(self):
        self.transport.makedirs(os.path.join(self.tmpdir, 'testdir'))
        children = self.transport.listdir(self.tmpdir)
        assert children == ['testdir']

    def test_recursive_makedirs(self):
        self.transport.makedirs(os.path.join(self.tmpdir, 'testdir/testdir2'))
        children = self.transport.listdir(self.tmpdir)
        assert children == ['testdir']

        nephews = self.transport.listdir(
            os.path.join(self.tmpdir, 'testdir'))
        assert nephews == ['testdir2']

    def test_open(self):
        fd = self.transport.open(os.path.join(self.tmpdir, 'testfile'), 'w+')
        fd.write("GC3")
        fd.close()

        fd = self.transport.open(os.path.join(self.tmpdir, 'testfile'), 'r')
        assert fd.read() == "GC3"

    def test_remove(self):
        fd = self.transport.open(os.path.join(self.tmpdir, 'testfile'), 'w+')
        fd.close()
        self.transport.remove(os.path.join(self.tmpdir, 'testfile'))
        assert self.transport.listdir(self.tmpdir) == []

    def test_chmod(self):
        remotefile = os.path.join(self.tmpdir, 'unauth')
        fd = self.transport.open(remotefile, 'w+')
        fd.close()

        self.transport.chmod(remotefile, 0000)

        try:
            fd = self.transport.open(remotefile, 'r')
        except TransportError:
            pass

    def test_get_and_put(self):
        with NamedTemporaryFile(prefix='gc3libs.test.',
                                suffix='.tmp',
                                mode='wt',
                                delete=False) as tmpfile:
            try:
                path = tmpfile.name  # save for later

                # create file with some content
                tmpfile.write("Test file")
                tmpfile.close()

                # copy it to the remote end using Transport.put()
                destfile = os.path.join(self.tmpdir, os.path.basename(path))
                self.transport.put(path, destfile)
                # remove the local file
                os.remove(path)

                # get the file using Transport.get()
                self.transport.get(destfile, path)

                # check the content
                assert open(path).read() == "Test file"

            finally:
                try:
                    os.unlink(path)
                except:
                    pass

    def test_open_failure_nonexistent_file(self):
        with pytest.raises(TransportError):
            # pylint: disable=invalid-name,unused-variable
            fd = self.transport.open(
                os.path.join(self.tmpdir, 'nonexistent'), 'r')

    def test_open_failure_unauthorized(self):
        # we cannot rely on *any* file being unreadable, as tests may
        # be running as `root` (e.g., in a Docker container), so we
        # need to explicitly create one
        with NamedTemporaryFile(prefix='gc3libs.test.',
                                suffix='.tmp',
                                mode='wt',
                                delete=False) as tmpfile:
            try:
                path = tmpfile.name  # save for later

                fd = tmpfile.fileno()
                os.fchmod(fd, 0o000)
                tmpfile.close()

                # now re-open with normal Python functions
                with pytest.raises(TransportError):
                    # pylint: disable=no-member
                    with self.transport.open(path, 'r') as stream:
                        assert stream is False
            finally:
                try:
                    os.unlink(path)
                except:
                    pass

    def test_remove_failure(self):
        with pytest.raises(TransportError):
            # pylint: disable=no-member
            self.transport.remove(
                os.path.join(self.tmpdir, 'nonexistent'))


class TestLocalTransport(StubForTestTransport):

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.transport = transport.LocalTransport()
        self.transport.connect()
        self.extra_setup()

@pytest.mark.skipif(
    'SshTransport' not in os.environ.get('GC3PIE_TESTS_ALLOW', ''),
    reason=("Skipping SSH test: SSH to localhost not allowed"
            " (set env variable `GC3PIE_TESTS_ALLOW` to `SshTransport` to run)"))
class TestSshTransport(StubForTestTransport):

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.transport = transport.SshTransport('localhost',
                                                ignore_ssh_host_keys=True)
        self.transport.connect()
        self.extra_setup()

# main: run tests

if __name__ == "__main__":
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, 2012,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

import datetime
import os
import shutil
import tempfile

import pytest

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.quantity import MB, seconds

State = gc3libs.Run.State

from faketransport import FakeTransport

files_to_remove = []


def correct_submit(jobid=123):
    out = """Your job %s ("DemoSGEApp") has been submitted
""" % jobid
    return (0, out, "")


def correct_qstat_queued(jobid=123):
    out = """  %s 0.00000 DemoSGEApp antonio      qw    03/15/2012 08:53:34 \
                                   1
""" % jobid
    return (0, out, "")


def correct_qstat_running(jobid=123):
    out = """  %s 0.55500 DemoSGEApp antonio      r     03/15/2012 08:53:42 \
all.q@compute-0-2.local            1
""" % jobid
    return (0, out, "")


def qacct_notfound(jobid=123):
    err = """error: job id %s not found
""" % jobid
    return (1, "", err)


def qdel_notfound(jobid=123):
    return (1, "",  """denied: job "%s" does not exist""" % jobid)


def qstat_notfound(jobid=123):
    return (1, "", "")


def correct_qacct_done(jobid=123):
    out = """
==============================================================
qname        all.q
hostname     compute-0-2.local
group        antonio
owner        antonio
project      NONE
department   defaultdepartment
jobname      DemoSGEApp
jobnumber    %s
taskid       undefined
account      sge
priority     0
qsub_time    Thu Mar 15 08:42:46 2012
start_time   Thu Mar 15 08:43:00 2012
end_time     Thu Mar 15 08:43:10 2012
granted_pe   NONE
slots        1
failed       0
exit_status  0
ru_wallclock 10
ru_utime     0.154
ru_stime     0.094
ru_maxrss    0
ru_ixrss     0
ru_ismrss    0
ru_idrss     0
ru_isrss     0
ru_minflt    22296
ru_majflt    0
ru_nswap     0
ru_inblock   0
ru_oublock   0
ru_msgsnd    0
ru_msgrcv    0
ru_nsignals  0
ru_nvcsw     306
ru_nivcsw    157
cpu          0.248
mem          0.000
io           0.006
iow          0.000
maxvmem      13.152M
arid         undefined
""" % jobid

    return (0, out, "")


def qsub_failed_jobnamestartswithdigit(jobname='123DemoSGEApp'):
    out = ""
    err = """Unable to run job: denied: "%s" is not a valid object name \
(cannot start with a digit).
Exiting.
""" % jobname
    return (1, out, err)

# def qsub_failed_acl():
#     out = ""
#     err = """qsub: Unauthorized Request  MSG=user ACL rejected the \
# submitting user: user amessina@argo.ictp.it, queue cm1
# """
#     return (159, out, err)


def qdel_success(jobid=123):
    return (0, "antonio has registered the job %s for deletion" % jobid, "")


def qdel_failed_acl(jobid=123):
    err = """antonio - you do not have the necessary privileges to delete \
the job "%s"
""" % jobid
    out = ""
    return (1, out, err)


class FakeApp(gc3libs.Application):

    def __init__(self, **extra_args):
        gc3libs.Application.__init__(
            self,
            arguments=['/bin/hostname'],  # mandatory
            inputs=[],                    # mandatory
            outputs=[],                   # mandatory
            output_dir="./fakedir",       # mandatory
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1, **extra_args)


class TestBackendSge(object):

    CONF = """
[resource/example]
type=sge
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64
queue=testing
enabled=True

[auth/ssh]
type=ssh
username=NONEXISTENT
"""

    @pytest.fixture(autouse=True)
    def setUp(self):
        (fd, self.tmpfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendSge.CONF)
        f.close()

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(self.tmpfile)

        self.core = gc3libs.core.Core(self.cfg)

        self.backend = self.core.get_backend('example')
        self.backend.transport = FakeTransport()
        self.transport = self.backend.transport
        # Basic responses
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['tracejob'] = qacct_notfound()
        self.transport.expected_answer['qdel'] = qdel_notfound()

        yield
        os.remove(self.tmpfile)

    def test_submission_failed(self):
        app = FakeApp()
        # Submission failed (jobname starting with a digit, cfr issue #250
        # at https://github.com/uzh/gc3pie/issues/250) This
        # first test will show the answer you would get if the job name
        # starts with a digit.
        self.transport.expected_answer['qsub'] = \
            qsub_failed_jobnamestartswithdigit()
        try:
            self.core.submit(app)
            assert False
        except Exception as e:
            assert isinstance(e, gc3libs.exceptions.LRMSError)
        assert app.execution.state == State.NEW

        # This second example will show how the Application.__init__()
        # method changes the jobname in order not to have a digit at the
        # beginning of it.
        app = FakeApp(jobname='123Demo')
        assert app.jobname == 'GC3Pie.123Demo'
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

    # Submission failed (unauthrozed user):
    #     t.expected_answer['qsub'] = qsub_failed_acl()
    #     try:
    #         self.core.submit(app)
    #     except Exception, e:
    #         assert isinstance(e, gc3libs.exceptions.LRMSError)
    #     assert_equal(app.execution.state, State.NEW)

    def test_sge_basic_workflow(self):
        app = FakeApp()
        # Succesful submission:
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

        # Update state. We would expect the job to be SUBMITTED
        self.transport.expected_answer['qstat'] = correct_qstat_queued()
        self.transport.expected_answer['qacct'] = qacct_notfound()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

        # Update state. We would expect the job to be RUNNING
        self.transport.expected_answer['qstat'] = correct_qstat_running()
        self.transport.expected_answer['qacct'] = qacct_notfound()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        # Job done. qstat doesn't find it, qacct should.
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['qacct'] = correct_qacct_done()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

    def test_qacct_parsing(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        self.transport.expected_answer['qacct'] = correct_qacct_done()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        job = app.execution
        # common job reporting values (see Issue 78)
        assert job.exitcode == 0
        assert job.returncode == 0
        assert job.duration == 10 * seconds
        assert (job.max_used_memory - 13.152 * MB) < 1 * MB  # floating approx
        assert job.used_cpu_time == 0.248 * seconds
        # SGE-specific values
        assert job.sge_queue == 'all.q'
        assert job.sge_jobname == 'DemoSGEApp'
        assert (job.sge_submission_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=15,
                                       hour=8,
                                       minute=42,
                                       second=46))
        assert (job.sge_start_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=15,
                                       hour=8,
                                       minute=43,
                                       second=0))
        assert (job.sge_completion_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=15,
                                       hour=8,
                                       minute=43,
                                       second=10))
        assert job.sge_failed == 0

    def test_delete_job(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)

        self.transport.expected_answer['qdel'] = qdel_success()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED

    def test_delete_job_twice(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

        self.transport.expected_answer['qdel'] = qdel_failed_acl()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED


def tearDownModule():
    for fname in files_to_remove:
        if os.path.isdir(fname):
            shutil.rmtree(fname)
        else:
            os.remove(fname)


def test_get_command():
    (fd, tmpfile) = tempfile.mkstemp()
    files_to_remove.append(tmpfile)
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=sge
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
qsub = /usr/local/bin/qsub -q testing
qacct = /usr/local/sbin/qacct
qstat = /usr/local/bin/qstat
qdel = /usr/local/bin/qdel # comments are ignored!
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    b = cfg.make_resources()['example']

    assert b.qsub == ['/usr/local/bin/qsub', '-q', 'testing']

    assert b._qacct == '/usr/local/sbin/qacct'
    assert b._qdel == '/usr/local/bin/qdel'
    assert b._qstat == '/usr/local/bin/qstat'


if __name__ == "__main__":
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/python

import csv
from datetime import datetime
import os
import sys

import gc3libs
from gc3libs.quantity import Duration, Memory
from gc3libs.session import Session


cwd = os.getcwd()

output = csv.writer(sys.stdout)

# header row
output.writerow([
    'inputfile',
    'funcname',
    'radius',
    'exitcode',
    'cores',
    'requested_memory',
    'used_memory',
    'duration',
    'started_at',
    'terminated_at',
])

for entry in os.listdir(cwd):
    if not entry[0].isupper():
        continue
    session = Session(os.path.join(cwd, entry))
    for task in session:
        # skip unfinished tasks
        if 'TERMINATED' not in task.execution.timestamp:
            continue
        try:
            parts = os.path.basename(task.output_dir).split('_')
            inputfile = parts[-2]
        except ValueError:
            sys.stderr.write("*** ERROR: Malformed jobname `%s` in task %s\n" % (task.output_dir, task))
            continue
        funcname = task.funcname
        radius = task.radius
        cores = task.requested_cores
        req_memory = task.requested_memory.amount(unit=Memory.MB)
        try:
            used_memory = task.execution.max_used_memory.amount(unit=Memory.MB)
        except AttributeError:
            used_memory = req_memory
        exitcode = task.execution._exitcode
        started_at = datetime.utcfromtimestamp(task.execution.timestamp['SUBMITTED'])
        terminated_at = datetime.utcfromtimestamp(task.execution.timestamp['TERMINATED'])
        try:
            duration = task.execution.duration.amount(unit=Duration.s)
        except AttributeError:
            duration = (terminated_at - started_at).total_seconds()
        output.writerow([
            inputfile,
            funcname,
            radius,
            exitcode,
            cores,
            req_memory,
            used_memory,
            duration,
            started_at.isoformat(),
            terminated_at.isoformat(),
        ])
    



#! /usr/bin/env python
#
#   greduction.py -- Front-end script for evaluating 'explain_matrix.py'
#   function over a large dataset.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-02-06:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import greduction
    greduction.GreductionScript().run()

import os
import sys
import time
import tempfile

import shutil
import json

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class
class GreductionApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'greduction'

    def __init__(self, language_file, **extra_args):

        inputs = dict()
        outputs = dict()

        inputs[language_file] = "./input.json"
        outputs['results/'] = os.path.join(extra_args['output_dir'],'results/')

        arguments = "python explain_matrix.py input.json"

        # check the optional inputs

        if extra_args.has_key('explain_matrix'):
            inputs[extra_args['explain_matrix']] = "./explain_matrix.py"

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'greduction.log',
            join=True,
            **extra_args)

class GreductionScript(SessionBasedScript):
    """
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GreductionApplication,
            stats_only_for = GreductionApplication,
            )

    def setup_options(self):
        self.add_param("-k", "--chunk", metavar="[NUM]", #type=executable_file,
                       dest="chunk_size", default="1000",
                       help="How to split the edges input data set.")

        self.add_param("-M", "--master", metavar="[PATH]",
                       dest="explain_matrix", default=None,
                       help="Location of 'explain_matrix.py' file.")

    def setup_args(self):

        self.add_param('language_file', type=str,
                       help="Input language file (in JSON format).")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        # check args:
        # XXX: make them position independent
        if not os.path.isfile(self.params.language_file):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid input language file: '%s'. File not found"
                % self.params.language_file)

        # Verify that 'self.params.chunk_size' is int
        int(self.params.chunk_size)
        self.params.chunk_size = int(self.params.chunk_size)

        if self.params.explain_matrix:
            assert os.path.isfile(self.params.explain_matrix)

    def new_tasks(self, extra):
        """
        """
        tasks = []

        for (input_file, index_chunk) in self._generate_chunked_files_and_list(self.params.language_file,
                                                                               self.params.chunk_size):


            jobname = "greduction-%s" % (str(index_chunk))

            extra_args = extra.copy()

            extra_args['index_chunk'] = str(index_chunk)

            extra_args['jobname'] = jobname

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        os.path.join('.computation',
                                                                                     jobname))
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        os.path.join('.computation',
                                                                                     jobname))


            if self.params.explain_matrix:
                extra_args['explain_matrix'] = self.params.explain_matrix

            self.log.debug("Creating Task for index : %d - %d" %
                           (index_chunk, (index_chunk + self.params.chunk_size)))

            tasks.append(GreductionApplication(
                    input_file,
                    **extra_args))

        return tasks

    def _generate_chunked_files_and_list(self, file_to_chunk, chunk_size):
        """
        Takes a file_name as input and a defined chunk size
        ( uses 1000 as default )
        returns a list of filenames 1 for each chunk created and a corresponding
        index reference
        """

        chunk = []
        chunk_files_dir = os.path.join(self.session.path,"tmp")

        # creating 'chunk_files_dir'
        if not(os.path.isdir(chunk_files_dir)):
            try:
                os.mkdir(chunk_files_dir)
            except OSError, osx:
                gc3libs.log.error("Failed while creating tmp folder %s. " % chunk_files_dir +
                                  "Error %s." % str(osx) +
                                  "Using default '/tmp'")
                chunk_files_dir = "/tmp"

        with open(file_to_chunk) as f:
            data = json.load(f)

        for index in range(0,len(data),chunk_size):
            (handle, tmp_filename) = tempfile.mkstemp(dir=chunk_files_dir,
                                                      prefix='greduction-',
                                                      suffix="%d.json" % index)
            with open(tmp_filename,'w') as fout:
                json.dump(dict(data.items()[index:index+chunk_size-1]),fout)
            chunk.append((tmp_filename,index))
        return chunk



#! /usr/bin/env python
#
#   gcrypto.py -- Front-end script for submitting multiple Crypto jobs to SMSCG.
"""
Front-end script for submitting multiple ``gnfs-cmd`` jobs to SMSCG.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcrypto --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2012-01-29:
    * Moved CryptoApplication from gc3libs.application

    * Restructured main script due to excessive size of initial
    jobs. SessionBaseScript generate a single SequentialTask.
    SequentialTask generated as many ParallelTasks as the whole range divided
    by the number of simultaneous active jobs.

    * Each ParallelTask lauches 'max_running' CryptoApplications
"""
__author__ = 'sergio.maffiolett@gc3.uzh.ch'
__docformat__ = 'reStructuredText'



# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gcrypto
    gcrypto.GCryptoScript().run()


# stdlib imports
import fnmatch
import logging
import os
import os.path
import sys
from pkg_resources import Requirement, resource_filename

# GC3Pie interface
import gc3libs
from gc3libs.cmdline import SessionBasedScript, existing_file, positive_int, nonnegative_int
from gc3libs import Application, Run, Task
import gc3libs.exceptions
import gc3libs.application
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection, ChunkedParameterSweep, RetryableTask

DEFAULT_INPUTFILE_LOCATION="srm://dpm.lhep.unibe.ch/dpm/lhep.unibe.ch/home/crypto/lacal_input_files.tgz"
DEFAULT_GNFS_LOCATION="srm://dpm.lhep.unibe.ch/dpm/lhep.unibe.ch/home/crypto/gnfs-cmd_20120406"

class CryptoApplication(gc3libs.Application):
    """
    Represent a ``gnfs-cmd`` job that examines the range `start` to `start+extent`.

    LACAL's ``gnfs-cmd`` invocation::

      $ gnfs-cmd begin length nth

    performs computations for a range: *begin* to *begin+length*,
    and *nth* is the number of threads spwaned.

    The following ranges are of interest: 800M-1200M and 2100M-2400M.

    CryptoApplication(param, step, input_files_archive, output_folder, **extra_args)
    """

    def __init__(self, start, extent, gnfs_location, input_files_archive, output, **extra_args):

        gnfs_executable_name = os.path.basename(gnfs_location)

        # # set some execution defaults...
        extra_args.setdefault('requested_cores', 4)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args['jobname'] = "LACAL_%s" % str(start + extent)
        extra_args['output_dir'] = os.path.join(extra_args['output_dir'], str(start + extent))
        extra_args['tags'] = [ 'APPS/CRYPTO/LACAL-1.0' ]
        extra_args['executables'] = ['./gnfs-cmd']
        extra_args['requested_memory'] = Memory(
            int(extra_args['requested_memory'].amount() / float(extra_args['requested_cores'])),
            unit=extra_args['requested_memory'].unit)

        gc3libs.Application.__init__(
            self,

            arguments = [ "./gnfs-cmd", start, extent, extra_args['requested_cores'], "input.tgz" ],
            inputs = {
                input_files_archive:"input.tgz",
                gnfs_location:"./gnfs-cmd",
                },
            outputs = [ '@output.list' ],
            # outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gcrypto.log',
            join=True,
            **extra_args
            )

    def terminated(self):
        """
        Checks whether the ``M*.gz`` files have been created.

        The exit status of the whole job is set to one of these values:

        *  0 -- all files processed successfully
        *  1 -- some files were *not* processed successfully
        *  2 -- no files processed successfully
        * 127 -- the ``gnfs-cmd`` application did not run at all.

        """
        # XXX: need to gather more info on how to post-process.
        # for the moment do nothing and report job's exit status

        if self.execution.exitcode:
            gc3libs.log.debug(
                'Application terminated. postprocessing with execution.exicode %d',
                self.execution.exitcode)
        else:
            gc3libs.log.debug(
                'Application terminated. No exitcode available')

        if self.execution.signal == 123:
            # XXX: this is fragile as it does not really applies to all
            # DataStaging errors.
            # Assume Data staging problem at the beginning of the job
            # resubmit
            self.execution.returncode = (0, 99)


class CryptoTask(RetryableTask):
    """
    Run ``gnfs-cmd`` on a given range
    """
    def __init__(self, start, extent, gnfs_location, input_files_archive, output, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            CryptoApplication(start, extent, gnfs_location, input_files_archive, output, **extra_args),
            # XXX: should decide which policy to use here for max_retries
            max_retries = 2,
            # keyword arguments
            **extra_args)

    def retry(self):
        """
        Resubmit a cryto application instance iff it exited with code 99.

        *Note:* There is currently no upper limit on the number of
        resubmissions!
        """
        if self.task.execution.exitcode == 99:
            return True
        else:
            return False


class CryptoChunkedParameterSweep(ChunkedParameterSweep):
    """
    Provided the beginning of the range `range_start`, the end of the
    range `range_end`, the slice size of each job `slice`,
    `CryptoChunkedParameterSweep` creates `chunk_size`
    `CryptoApplication`s to be executed in parallel.

    Every update cycle it will check how many new CryptoApplication
    will have to be created (each of the launching in parallel
    DEFAULT_PARALLEL_RANGE_INCREMENT CryptoApplications) as the
    following rule: [ (end-range - begin_range) / step ] /
    DEFAULT_PARALLEL_RANGE_INCREMENT
    """


    def __init__(self, range_start, range_end, slice, chunk_size,
                 input_files_archive, gnfs_location, output_folder, **extra_args):

        # remember for later
        self.range_start = range_start
        self.range_end = range_end
        self.parameter_count_increment = slice * chunk_size
        self.input_files_archive = input_files_archive
        self.gnfs_location = gnfs_location
        self.output_folder = output_folder
        self.extra_args = extra_args

        ChunkedParameterSweep.__init__(
            self, range_start, range_end, slice, chunk_size, **self.extra_args)

    def new_task(self, param, **extra_args):
        """
        Create a new `CryptoApplication` for computing the range
        `param` to `param+self.parameter_count_increment`.
        """
        return CryptoTask(param, self.step, self.gnfs_location, self.input_files_archive, self.output_folder, **self.extra_args.copy())




## the script itself

class GCryptoScript(SessionBasedScript):
    # this will be display as the scripts' `--help` text
    """
Like a `for`-loop, the ``gcrypto`` driver script takes as input
three mandatory arguments:

1. RANGE_START: initial value of the range (e.g., 800000000)
2. RANGE_END: final value of the range (e.g., 1200000000)
3. SLICE: extent of the range that will be examined by a single job (e.g., 1000)

For example::

  gcrypto 800000000 1200000000 1000

will produce 400000 jobs; the first job will perform calculations
on the range 800000000 to 800000000+1000, the 2nd one will do the
range 800001000 to 800002000, and so on.

Inputfile archive location (e.g. lfc://lfc.smscg.ch/crypto/lacal/input.tgz)
can be specified with the '-i' option. Otherwise a default filename
'input.tgz' will be searched in current directory.

Job progress is monitored and, when a job is done,
output is retrieved back to submitting host in folders named:
'range_start + (slice * actual step)'

The `gcrypto` command keeps a record of jobs (submitted, executed and
pending) in a session file (set name with the '-s' option); at each
invocation of the command, the status of all recorded jobs is updated,
output from finished jobs is collected, and a summary table of all
known jobs is printed.  New jobs are added to the session if new input
files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; `gcrypto` will delay submission
of newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            stats_only_for = CryptoApplication,
            )


    def setup_args(self):
        """
        Set up command-line argument parsing.

        The default command line parsing considers every argument as
        an (input) path name; processing of the given path names is
        done in `parse_args`:meth:
        """

        # self.add_param('args',
        #                nargs='*',
        #                metavar=
        #                """
        #                [range_start] [range_end] [slice],
        #                help=[range_start]: Positive integer value of the range start.
        #                [range_end]: Positive integer value of the range end.
        #                [slice]: Positive integer value of the increment.
        #                """
        #                )


        self.add_param('range_start', type=nonnegative_int,
                  help="Non-negative integer value of the range start.")
        self.add_param('range_end', type=positive_int,
                  help="Positive integer value of the range end.")
        self.add_param('slice', type=positive_int,
                  help="Positive integer value of the increment.")

    def parse_args(self):
        # XXX: why is this necessary ? shouldn't add_params of 'args' handle this ?
        # check on the use of nargs and type.
        # if len(self.params.args) != 3:
        #     raise ValueError("gcrypto takes exaclty 3 arguments (%d are given)" % len(self.params.args))
        # self.params.range_start = int(self.params.args[0])
        # self.params.range_end = int(self.params.args[1])
        # self.params.slice = int(self.params.args[2])

        if self.params.range_end <= self.params.range_start:
            # Failed
            raise ValueError("End range cannot be smaller than Start range. Start range %d. End range %d" % (self.params.range_start, self.params.range_end))

    def setup_options(self):
        self.add_param("-i", "--input-files", metavar="PATH",
                       action="store", dest="input_files_archive",
                       default=DEFAULT_INPUTFILE_LOCATION,
                       help="Path to the input files archive."
                       " By default, the preloaded input archive available on"
                       " SMSCG Storage Element will be used: "
                       " %s" % DEFAULT_INPUTFILE_LOCATION)
        self.add_param("-g", "--gnfs-cmd", metavar="PATH",
                       action="store", dest="gnfs_location",
                       default=DEFAULT_GNFS_LOCATION,
                       help="Path to the executable script (gnfs-cmd)"
                       " By default, the preloaded gnfs-cmd available on"
                       " SMSCG Storage Element will be used: "
                       " %s" % DEFAULT_GNFS_LOCATION)


    def new_tasks(self, extra):
        yield (
            "%s-%s" % (str(self.params.range_start),str(self.params.range_end)), # jobname
            CryptoChunkedParameterSweep,
            [ # parameters passed to the constructor, see `CryptoSequence.__init__`
                self.params.range_start,
                self.params.range_end,
                self.params.slice,
                self.params.max_running, # increment of each ParallelTask
                self.params.input_files_archive, # path to input.tgz
                self.params.gnfs_location, # path to gnfs-cmd
                self.params.output, # output folder
                ],
            extra.copy()
            )


    def before_main_loop(self):
        """
        Ensure each instance of `ChunkedParameterSweep` has
        `chunk_size` set to the maximum allowed number of jobs.
        """
        for task in self.session:
            assert isinstance(task, CryptoChunkedParameterSweep)
            task.chunk_size = self.params.max_running



#! /usr/bin/env python

"""
Interface to different resource management systems for the GC3Libs.
"""

# Copyright (C) 2009-2015  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'


from functools import wraps

import gc3libs
import gc3libs.exceptions
from gc3libs.quantity import Memory
from gc3libs.quantity import Duration
import gc3libs.utils


class LRMS(gc3libs.utils.Struct):

    """Base class for interfacing with a computing resource.

    The following construction parameters are also set as instance
    attributes.  All of them are mandatory, except `auth`.

    +---------------------+--------------+-------------------------------+
    |Attribute name       |Expected Type |Meaning                        |
    +=====================+==============+===============================+
    |`name`               |string        |A unique identifier for this   |
    |                     |              |resource, used for generating  |
    |                     |              |error message.                 |
    +---------------------+--------------+-------------------------------+
    |`architecture`       |set of        |Should contain one entry per   |
    |                     |`Run.Arch`    |each architecture              |
    |                     |values        |supported. Valid architecture  |
    |                     |              |values are constants in the    |
    |                     |              |`gc3libs.Run.Arch` class.      |
    +---------------------+--------------+-------------------------------+
    |`auth`               |string        |A `gc3libs.authentication.Auth`|
    |                     |              |instance that will be used to  |
    |                     |              |access the computational       |
    |                     |              |resource associated with this  |
    |                     |              |backend.  The default value    |
    |                     |              |`None` is used to mean that no |
    |                     |              |authentication credentials are |
    |                     |              |needed (e.g., access to the    |
    |                     |              |resource has been              |
    |                     |              |pre-authenticated) or is       |
    |                     |              |managed outside of GC3Pie).    |
    +---------------------+--------------+-------------------------------+
    |`max_cores`          |int           |Maximum number of CPU cores    |
    |                     |              |that GC3Pie can allocate on    |
    |                     |              |this resource.                 |
    +---------------------+--------------+-------------------------------+
    |`max_cores_per_job`  |int           |Maximum number of CPU cores    |
    |                     |              |that GC3Pie can allocate on    |
    |                     |              |this resource *for a single    |
    |                     |              |job*.                          |
    +---------------------+--------------+-------------------------------+
    |`max_memory_per_core`|Memory        |Maximum memory that            |
    |                     |              |GC3Pie can allocate to jobs on |
    |                     |              |this resource.  The value is   |
    |                     |              |*per core*, so the actual      |
    |                     |              |amount allocated to a single   |
    |                     |              |job is the value of this entry |
    |                     |              |multiplied by the number of    |
    |                     |              |cores requested by the job.    |
    +---------------------+--------------+-------------------------------+
    |`max_walltime`       |Duration      |Maximum wall-clock time        |
    |                     |              |that can be allotted           |
    |                     |              |to a single job running on this|
    |                     |              |resource.                      |
    +---------------------+--------------+-------------------------------+

    The above should be considered *immutable* attributes: they are
    specified at construction time and changed never after.

    The following attributes are instead dynamically provided (i.e.,
    defined by the `get_resource_status()` method or similar), thus
    can change over the lifetime of the object:

    ===================  =====
    Attribute name       Type
    ===================  =====
    free_slots           int
    user_run             int
    user_queued          int
    queued               int
    ===================  =====

    """

    def __init__(self, name,
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime, auth=None,
                 # additional arguments can set instance attributes
                 **extra_args):
        gc3libs.utils.Struct.__init__(self, **extra_args)

        self.name = str(name)
        self.updated = False

        if len(architecture) == 0:
            raise gc3libs.exceptions.InvalidType(
                "Empty value list for mandatory attribute 'architecture'")
        self.architecture = architecture

        self.max_cores = int(max_cores)
        self.max_cores_per_job = int(max_cores_per_job)
        assert isinstance(max_memory_per_core, Memory), \
            ("Expected `Memory` value for `max_memory_per_core`, "
             "got %s instead." % (type(max_memory_per_core),))
        self.max_memory_per_core = max_memory_per_core
        assert isinstance(max_walltime, Duration), \
            ("Expected `Duration` value for `max_walltime`, got %s instead."
             % (type(max_walltime),))
        self.max_walltime = max_walltime

        # see `authenticated` below
        self._auth_fn = auth

        gc3libs.log.info(
            "Computational resource '%s' initialized successfully.", self.name)

    @staticmethod
    def authenticated(fn):
        """
        Decorator: mark a function as requiring authentication.

        Each invocation of the decorated function causes a call to the
        `get` method of the authentication object (configured with the
        `auth` parameter to the class constructor).
        """
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self._auth_fn is not None:
                try:
                    self._auth_fn()
                except Exception as err:
                    # log error and disable resource - if the "*Auth"
                    # object cannot be instanciated, there is not much
                    # we can do...
                    gc3libs.log.warning(
                        "Problems initializing authentication backend"
                        " for resource '%s': %s.", self.name, err)
                    gc3libs.log.error(
                        "Resource '%s' will be disabled"
                        " because of authentication problems.",
                        self.name)
                    self.enabled = False
            return fn(self, *args, **kwargs)
        return wrapper

    def cancel_job(self, app):
        """
        Cancel a running job.  If `app` is associated to a queued or
        running remote job, tell the execution middleware to cancel
        it.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.cancel_job()` called "
            "- this should have been defined in a derived class.")

    def free(self, app):
        """
        Free up any remote resources used for the execution of `app`.
        In particular, this should delete any remote directories and
        files.

        Call this method when `app.execution.state` is anything other
        than `TERMINATED` results in undefined behavior and will
        likely be the cause of errors later on.  Be cautious.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.free()` called "
            "- this should have been defined in a derived class.")

    def get_resource_status(self):
        """
        Update the status of the resource associated with this `LRMS`
        instance in-place.  Return updated `Resource` object.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.get_resource_status()` called "
            "- this should have been defined in a derived class.")

    def get_results(self, job, download_dir,
                    overwrite=False, changed_only=True):
        """
        Retrieve job output files into local directory `download_dir`.

        Directory `download_dir` must already exists.

        If optional 3rd argument `overwrite` is ``False`` (default),
        then existing files within `download_dir` (or subdirectories
        thereof) will *not* be altered in any way.

        If `overwrite` is instead ``True``, then the (optional) 4th
        argument `changed_only` determines what files are overwritten:

        - if `changed_only` is ``True`` (default), then only files for
          which the source has a different size or has been modified
          more recently than the destination are copied;

        - if `changed_only` is ``False``, then *all* files in `source`
          will be copied into `destination`, unconditionally.

        Output files that do not exist in `download_dir` will be
        copied, independently of the `overwrite` and `changed_only`
        settings.

        :param Task job:
          the `Task`:class: instance whose output should be retrieved
        :param str download_dir:
          path to download files into
        :param bool overwrite:
          if `False`, do not download files that already exist
        :param bool changed_only:
          if both this and `overwrite` are `True`, only overwrite
          those files such that the source is newer or different in
          size than the destination.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.get_results()` called "
            "- this should have been defined in a derived class.")

    def update_job_state(self, app):
        """
        Query the state of the remote job associated with `app` and
        update `app.execution.state` accordingly.  Return the
        corresponding `Run.State`; see `Run.State` for more details.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.update_state()` called "
            "- this should have been defined in a derived class.")

    def submit_job(self, application, job):
        """
        Submit an `Application` instance to the configured
        computational resource; return a `gc3libs.Job` instance for
        controlling the submitted job.

        This method only returns if the job is successfully submitted;
        upon any failure, an exception is raised.

        *Note:*

          1. `job.state` is *not* altered; it is the caller's
             responsibility to update it.

          2. the `job` object may be updated with any information that
             is necessary for this LRMS to perform further operations on it.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.submit_job()` called "
            "- this should have been defined in a derived class.")

    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        """
        Download `size` bytes (at offset `offset` from the start) from
        remote file `remote_filename` and write them into
        `local_file`.  If `size` is `None` (default), then snarf
        contents of remote file from `offset` unto the end.

        First argument `remote_filename` is the path to a file
        relative to the remote job "sandbox".

        Argument `local_file` is either a local path name (string), or
        a file-like object supporting a `.write()` method.  If
        `local_file` is a path name, it is created if not existent,
        otherwise overwritten.  In any case, upon exit from this
        procedure, the stream will be positioned just after the
        written bytes.

        Fourth optional argument `offset` is the offset from the start
        of the file.  If `offset` is negative, it is interpreted as an
        offset from the *end* of the remote file.

        Any exception raised by operations will be re-raised to the caller.
        """
        raise NotImplementedError(
            "Abstract method `LRMS.peek()` called "
            "- this should have been defined in a derived class.")

    def validate_data(self, data_file_list=None):
        """
        Return True if the list of files is expressed in one of the
        file transfer protocols the LRMS supports.

        Return False otherwise.
        """
        raise NotImplementedError(
            "Abstract method 'LRMS.validate_data()' called "
            "- this should have been defined in a derived class.")

    def close(self):
        """
        Implement gracefully close on LRMS dependent resources
        e.g. transport
        """
        raise NotImplementedError(
            "Abstract method 'LRMS.close()' called "
            "- this should have been defined in a derived class.")


# main: run tests
if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gcombi.py -- Front-end script for evaluating Matlab functions
#   function over a large number of parameters.
#
#   Copyright (C) 2015, 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation, either version 3 of
#   the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple Matlab jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcombi.py --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-04-20:
  * Initial version
"""
__author__ = 'Tyanko Aleksiev <tyanko.aleksiev@uzh.ch>'
__docformat__ = 'reStructuredText'


if __name__ == "__main__":
    import gcombi
    gcombi.GCombiScript().run()

import os
import sys
import time
import tempfile

import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class
class GCombiApplication(Application):
    """
    Custom class to for the execution of three matlab models used by Gaia Lombardi at from
    the Department of Economics: project fehr.econ.uzh.
    Each model takes as input two parameters: 'phenotype' and 'chromosome' and this is the work-flow:

    model1 phenotype chromosome -> outout1
    model2 phenotype output1 -> output2
    model3 phenotype output2 -> output3.

    The current implementation of the three models requires a submission like this:

    model1 phenotype chromosome
    model2 phenotype chromosome
    model3 phenotype chromosome

    The project is one shot and has to execute 96 phenotypes each against 22 chromosoms for a total of
    ~21000 cpu hours. More than 4GB are needed for single execution so at least a 2cpu-8ram-hpc flavor
    has to be used.
    """
    application_name = 'gcombi'

    def __init__(self, input_phenotype, input_chromosom, **extra_args):

        inputs = dict()
        outputs = dict()

        output_dir = "./results"
        outputs[output_dir] = output_dir

        # execution wrapper needs to be added anyway
        gcombi_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gcombi.sh")
        inputs[gcombi_wrapper_sh] = os.path.basename(gcombi_wrapper_sh)
        inputs[input_phenotype] = os.path.basename(input_phenotype)
        inputs[input_chromosom] = os.path.basename(input_chromosom)

        command = "./%s ./%s ./%s" % (os.path.basename(gcombi_wrapper_sh), os.path.basename(input_phenotype), os.path.basename(input_chromosom))

        Application.__init__(
            self,
            arguments = command,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gcombi.log',
            join=True,
            executables = "./%s" % os.path.basename(gcombi_wrapper_sh),
            **extra_args)


class GCombiScript(SessionBasedScript):
    """
    The ``gcombi`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gcombi`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gcombi``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GCombiApplication,
            stats_only_for = GCombiApplication,
            )

    def setup_args(self):
        self.add_param('phenotypes_dir', type=str, help="Specify the directory containing the phenotypes")
        self.add_param('chromosomes_dir', type=str, help="Specify the directory containing the cromosomes for that phenotype")

    def parse_args(self):
        """
        Check presence of input folder (should contains matlab scripts).
        path to command_file should also be valid.
        """

    def new_tasks(self, extra):
        """
        """
        tasks = []
        allchromosoms = ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22']

        for phenotype in os.listdir(self.params.phenotypes_dir):
            jobname = phenotype.split(".")[0]
            phenotype_input_file = self.params.phenotypes_dir + os.path.basename(phenotype)

            for chromosom in os.listdir(self.params.chromosomes_dir):
                chromosom_input_file = self.params.chromosomes_dir + os.path.basename(chromosom)
                jobname = "gcombi-%s-%s" % (os.path.basename(phenotype).split(".")[0], os.path.basename(chromosom).split(".")[0])

                extra_args = extra.copy()

                extra_args['jobname'] = jobname

                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

                self.log.debug("Creating Application for parameters : %s %s" %
                               (phenotype, chromosom))

                tasks.append(GCombiApplication(
                    phenotype_input_file,
                    chromosom_input_file,
                    **extra_args))

        return tasks



#! /usr/bin/env python
#
"""
Accessors for object attributes and container items.
"""
# Copyright (C) 2011-2012, 2019  University of Zurich.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'


from gc3libs.utils import getattr_nested


# tag object for catching the "no value passed" in `GetAttr` and
# `value_at_index` (cannot use `None` as it's a legit value!)
_none = object()


class GetValue(object):

    """
    Provide easier compositional syntax for `GetAttributeValue` and
    `GetItemValue`.

    Instances of `GetAttributeValue` and `GetItemValue` can be composed by
    passing one as `xform` parameter to the other; however, this
    results in the writing order being the opposite of the composition
    order: for instance, to create an accessor to evaluate `x.a[0]`
    for any Python object `x`, one has to write::

       >>> from gc3libs import Struct
       >>> fn1 = GetItemValue(0, GetAttributeValue('a'))

    The `GetValue` class allows to write accessor expressions the way
    they are normally written in Python::

       >>> GET = GetValue()
       >>> fn2 = GET.a[0]
       >>> x = Struct(a=[21,42], b='foo')
       >>> fn1(x)
       21
       >>> fn2(x)
       21

    The optional `default` argument specifies a value that should be
    used in case the required attribute or item is not found:

       >>> fn3 = GetValue(default='no value found').a[3]
       >>> fn3(x) == 'no value found'
       True

    """
    __slots__ = ('default',)

    def __init__(self, default=_none):
        self.default = default

    def __call__(self, obj):
        # identity accessor
        return obj

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(
                "'%s' object has no attribute '%s'"
                % (self.__class__.__name__, name))
        return GetAttributeValue(
            name, xform=(lambda obj: self(obj)), default=self.default)

    def __getitem__(self, place):
        return GetItemValue(
            place, xform=(lambda obj: self(obj)), default=self.default)

    def ONLY(self, specifier):
        """
        Restrict the action of the accessor expression to members of a certain
        class; return default value otherwise.

        The invocation to `only`:meth: should *always be last*::

            >>> from gc3libs import Struct
            >>> fn = GetValue(default='foo').a[0].ONLY(Struct)
            >>> fn(Struct(a=['bar','baz'])) == 'bar'
            True
            >>> fn(dict(a=['bar','baz'])) == 'foo'
            True

        If it's not last, you will get `AttributeError` like the following:

            >>> fn = GetValue().ONLY(Struct).a[0]
            >>> fn(dict(a=[0,1]))
            Traceback (most recent call last):
              ...
            AttributeError: 'NoneType' object has no attribute 'a'

        """
        return GetOnly(
            specifier, xform=(lambda obj: self(obj)), default=self.default)


GET = GetValue()
"""
Constant identity getter.

Use this for better readability (e.g., `GET[0]` instead of
`GetValue()[0]`).
"""


class GetAttributeValue(GetValue):

    """
    Return an accessor function for the given attribute.

    An instance of `GetAttributeValue` is a callable that, given any
    object, returns the value of its attribute `attr`, whose name is
    specified in the `GetAttributeValue` constructor::

       >>> from gc3libs import Struct
        >>> fn = GetAttributeValue('x')
        >>> a = Struct(x=1, y=2)
        >>> fn(a)
        1

    The accessor raises `AttributeError` if no such attribute
    exists)::

        >>> b = Struct(z=3)
        >>> fn(b)
        Traceback (most recent call last):
           ...
        AttributeError: 'Struct' object has no attribute 'x'

    However, you can specify a default value, in which case the
    default value is returned and no error is raised::

        >>> fn = GetAttributeValue('x', default=42)
        >>> fn(b)
        42
        >>> fn = GetAttributeValue('y', default=None)
        >>> print(fn(b))
        None

    In other words, if `fn = GetAttributeValue('x')`, then `fn(obj)`
    evaluates to `obj.x`.

    If the string `attr` contains any dots, then attribute lookups are
    chained: if `fn = GetAttributeValue('x.y')` then `fn(obj)` evaluates to
    `obj.x.y`::

        >>> fn = GetAttributeValue('x.y')
        >>> a = Struct(x=Struct(y=42))
        >>> fn(a)
        42

    The optional second argument `xform` allows composing the accessor
    with an arbitrary function that is passed an object and should
    return a (possibly different) object whose attributes should be
    looked up.  In other words, if `xform` is specified, then the
    returned accessor function computes `xform(obj).attr` instead of
    `obj.attr`.

    This allows combining `GetAttributeValue` with `GetItemValue`:meth: (which
    see), to access objects in deeply-nested data structures; see
    `GetItemValue`:class: for examples.

    """
    __slots__ = ('attr', 'xform', 'default')

    def __init__(self, attr, xform=(lambda obj: obj), default=_none):
        self.attr = attr
        self.xform = xform
        self.default = default

    def __call__(self, obj):
        try:
            return getattr_nested(self.xform(obj), self.attr)
        except AttributeError:
            if self.default is not _none:
                return self.default
            else:
                raise


class GetItemValue(GetValue):

    """
    Return accessor function for the given item in a sequence.

    An instance of `GetItemValue` is a callable that, given any
    sequence/container object, returns the value of the item at its
    place `idx`::

        >>> fn = GetItemValue(1)
        >>> a = 'abc'
        >>> fn(a) == 'b'
        True
        >>> b = { 1:'x', 2:'y' }
        >>> fn(b) == 'x'
        True

    In other words, if `fn = GetItemValue(x)`, then `fn(obj)` evaluates
    to `obj[x]`.

    Note that the returned function `fn` raises `IndexError` or `KeyError`,
    (depending on the type of sequence/container) if place `idx` does not
    exist::

        >>> fn = GetItemValue(42)
        >>> a = list('abc')
        >>> fn(a)
        Traceback (most recent call last):
           ...
        IndexError: list index out of range
        >>> b = dict(x=1, y=2, z=3)
        >>> fn(b)
        Traceback (most recent call last):
           ...
        KeyError: 42

    However, you can specify a default value, in which case the
    default value is returned and no error is raised::

        >>> fn = GetItemValue(42, default='foo')
        >>> fn(a) == 'foo'
        True
        >>> fn(b) == 'foo'
        True

    The optional second argument `xform` allows composing the accessor
    with an arbitrary function that is passed an object and should
    return a (possibly different) object where the item lookup should
    be performed.  In other words, if `xform` is specified, then the
    returned accessor function computes `xform(obj)[idx]` instead of
    `obj[idx]`.  For example::

        >>> c = 'abc'
        >>> fn = GetItemValue(1, xform=(lambda s: s.upper()))
        >>> fn(c) == 'B'
        True

        >>> c = (('a',1), ('b',2))
        >>> fn = GetItemValue('a', xform=dict)
        >>> fn(c)
        1


    This allows combining `GetItemValue` with `GetAttrValue`:class:
    (which see), to access objects in deeply-nested data structures.
    """
    __slots__ = ('idx', 'xform', 'default')

    def __init__(self, place, xform=(lambda obj: obj), default=_none):
        self.idx = place
        self.xform = xform
        self.default = default

    def __call__(self, obj):
        try:
            return self.xform(obj)[self.idx]
        except (KeyError, IndexError):
            if self.default is not _none:
                return self.default
            else:
                raise


class GetOnly(GetValue):

    """
    Apply accessor function to members of a certain class; return a
    default value otherwise.

    The `GetOnly` accessor performs just like `GetValue`, but is
    effective only on instances of a certain class; if the accessor
    function is passed an instance of a different class, the default
    value is returned::

       >>> from gc3libs import Struct
       >>> fn4 = GetOnly(Struct, default=42)
       >>> isinstance(fn4(Struct(foo='bar')), Struct)
       True
       >>> isinstance(fn4(dict(foo='bar')), dict)
       False
       >>> fn4(dict(foo='bar'))
       42

    If `default` is not specified, then `None` is returned::

       >>> fn5 = GetOnly(Struct)
       >>> repr(fn5(dict(foo='bar')))
       'None'

    """
    __slots__ = ('only', 'xform', 'default')

    def __init__(self, only, xform=(lambda obj: obj), default=_none):
        self.only = only
        self.xform = xform
        self.default = default

    def __call__(self, obj):
        # only apply `xform` if `obj` matches a class in `only`
        if isinstance(obj, self.only):
            return self.xform(obj)
        else:
            if self.default is not _none:
                return self.default
            else:
                return None


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="get",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gmodis.py -- Front-end script for submitting multiple `MODIS` Matlab-based jobs.
#
#   Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple Matlab function jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gmodis.py --help`` for program usage instructions.

Input parameters consists of:
:param str input_dir: Path to the folder containing all the input data to be
                      processed.
       example: FSC10A1.A2000083_aT_fsc_stitch.mat

Option paramenters consist of:
:param str fsc_dir: Path to the folder containing all the FSC input files necessary
                    to process each of the input data.
                    Note: on a cloud based solution, the FSC input folder
                          will be re-deployed on the reference appliances.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2013-12-30:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gmodis
    gmodis.GmodisScript().run()

import os
import sys
import time
import tempfile

import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## custom application class
class GmodisApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gmodis'

    def __init__(self, input_file, **extra_args):

        # setup output references
        self.output_dir = extra_args['output_dir']

        # setup input references

        inputs = dict()

        gmodis_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gmodis_wrapper.sh")
        inputs[gmodis_wrapper_sh] = os.path.basename(gmodis_wrapper_sh)

        _command = []

        _command.append("./%s" % os.path.basename(gmodis_wrapper_sh))

        # Add denug info
        _command.append("-d")

        if extra_args.has_key('fsc_dir'):
            inputs.update(dict((os.path.join(extra_args['fsc_dir'],v),
                            os.path.join(
                                os.path.basename(extra_args['fsc_dir']),
                                         v))
                           for v in os.listdir(extra_args['fsc_dir'])))




            _command.append("-f ./%s " % os.path.basename(extra_args['fsc_dir']))

        if extra_args.has_key('gmodis_funct'):
            # e.g. ('/home/data/matlab/gmodis','~/bin/gmodis')
            inputs[extra_args['gmodis_funct']] = os.path.basename(extra_args['gmodis_funct'])

            _command.append("-x ./%s " % os.path.basename(extra_args['gmodis_funct']))

        if extra_args.has_key('matlab_driver'):
            inputs[extra_args['matlab_driver']] = os.path.basename(extra_args['matlab_driver'])

            _command.append("-s ./%s " % os.path.basename(extra_args['matlab_driver']))

        inputs[input_file] = os.path.basename(input_file)
        _command.append(os.path.basename(input_file))

        outputs =  gc3libs.ANY_OUTPUT

        # Add memory requirement
        extra_args['requested_memory'] = 16*GB

        Application.__init__(
            self,
            arguments = _command,
            executables = "./%s" % os.path.basename(gmodis_wrapper_sh),
            inputs = inputs,
            outputs = outputs,
            stdout = 'gmodis.log',
            join=True,
            **extra_args)

class GmodisTask(RetryableTask):
    def __init__(self, input_file, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GmodisApplication(
                input_file,
                **extra_args),
            **extra_args
            )

    def retry(self):
        """
        Task will be retried iif the application crashed
        due to an error within the exeuction environment
        (e.g. VM crash or LRMS kill)
        """
        # XXX: check whether it is possible to distinguish
        # between the error conditions and set meaningfull exitcode
        return False


class GmodisScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``gmodis`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gmodis`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GmodisApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GmodisApplication,
            )

    def setup_options(self):
        self.add_param("-f", "--fsc", metavar="PATH",
                       dest="fsc_dir", default=None,
                       help="Path to the FSC data folder.")

        self.add_param("-x", "--executable", metavar="PATH",
                       dest="gmodis_funct", default=None,
                       help="gmodis binary function.")

        self.add_param("-D", "--driver", metavar="PATH",
                       dest="matlab_driver", default=None,
                       help="Path to alternative Matlab driver "+
                       "script.")

    def setup_args(self):

        self.add_param('input_dir', type=str,
                       help="Path to input folder.")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        # check args:
        # XXX: make them position independent
        if not os.path.isdir(self.params.input_dir):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to input folder: '%s'. Path not found"
                % self.params.input_dir)

        self.log.info("Input dir: %s" % self.params.input_dir)

        if self.params.fsc_dir:
            if not os.path.isdir(self.params.fsc_dir):
                raise gc3libs.exceptions.InvalidUsage(
                    "Input FSC folder '%s' does not exists"
                    % self.params.input_dir)
            self.log.info("Input data dir: [%s]" % self.params.fsc_dir)
        else:
            self.log.info("Input data dir: [use remote]")

        if self.params.gmodis_funct:
            if not os.path.isfile(self.params.gmodis_funct):
                raise gc3libs.exceptions.InvalidUsage(
                    "gmodis binary '%s' does not exists"
                    % self.params.gmodis_funct)
            self.log.info("gmodis binary: [%s]" % self.params.gmodis_funct)
        else:
            self.log.info("gmodis binary: [use remote]")


    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """


        tasks = []

        for input_file in os.listdir(os.path.abspath(self.params.input_dir)):
            try:
                # Take only .mat files
                if input_file.endswith('.mat'):

                    # Use first sequence in input file name as jobname
                    # e.g. FSC10A1.A2000083_aT_fsc_stitch.mat will get
                    # gmodis-FSC10A1.A2000083
                    jobname = "gmodis-%s" % input_file[:16]

                    extra_args = extra.copy()
                    extra_args['jobname'] = jobname
                    # FIXME: ignore SessionBasedScript feature of customizing
                    # output folder
                    extra_args['output_dir'] = self.params.output

                    extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
                    extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
                    extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
                    extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

                    self.log.info("Creating Task for input file: %s" % input_file)

                    if self.params.matlab_driver:
                        extra_args['matlab_driver'] = self.params.matlab_driver

                    if self.params.fsc_dir:
                        extra_args['fsc_dir'] = self.params.fsc_dir

                    if self.params.gmodis_funct:
                        extra_args['gmodis_funct'] = self.params.gmodis_funct

                    tasks.append(GmodisTask(
                        os.path.join(os.path.abspath(self.params.input_dir),input_file),
                        **extra_args))

            except Exception, ex:
                self.log.error("Unexpected error. Error type: %s, Message: %s" % (type(ex),str(ex)))

        return tasks



#! /usr/bin/env python
#
#   gfdiv.py -- GC3Pie front-end for running the
#   "functional diversity" code by F. D. Schneider
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for parallelizing execution of the "functional
diversity" calculations.

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gfdiv --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
* 2016-05-18: Allow repeating set of arguments to form
              a composite session.
* 2016-04-22: Changed 2nd positional arg `values` to
              accept a *set* of `nHood` values to run.
              Also, allow setting the actual arguments used
              in the MATLAB function call.
* 2016-04-12: Do *not* force MATLAB to run single-threaded.
* 2016-02-24: Retry each task upon "out of memory" errors.
* 2016-01-25: Initial release.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import gfdiv
    gfdiv.GfdivScript().run()


# std module imports
from argparse import ArgumentError
import itertools
import os
from os.path import basename, exists, join, realpath
import sys
from pkg_resources import Requirement, resource_filename

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, existing_file
from gc3libs.quantity import Memory, kB, MB, GB, Duration, days, hours, minutes, seconds
from gc3libs.utils import basename_sans, fgrep, irange, occurs, parse_range
from gc3libs.workflow import RetryableTask


## custom application class

class MatlabRetryOnOutOfMemory(RetryableTask):
    """
    Run a task and retry it if failed with an out-of-memory condition.

    Memory requirements will be increased each time by a configurable
    amount until a set maximum is reached; when that happens, we just
    give up retrying and mark the task as failed.

    Proper detection of the out-of-memory condition relies on MATLAB's
    error message ``Out of memory.`` being detected in the task's
    STDOUT stream.
    """

    def __init__(self, task, increment=1*GB, maximum=31*GB, **extra_args):
        self.increment = increment
        self.maximum = maximum
        RetryableTask.__init__(self, task, **extra_args)

    def retry(self):
        last_run = self.task.execution
        if last_run.returncode == 0:
            gc3libs.log.debug(
                "%s: Task finished successfully,"
                " *not* resubmitting it again.", self.task)
            return False
        else:
            # task errored out, find out why
            try:
                requested_memory = self.task.requested_memory or last_run.max_used_memory
            except AttributeError:
                requested_memory = last_run.max_used_memory
            generic_memory_error = (last_run.max_used_memory > requested_memory)
            task_stderr = os.path.join(self.task.output_dir,
                                       self.task.stdout if self.task.join else self.task.stderr)
            matlab_memory_error = (
                occurs('Out of memory.', task_stderr, fgrep)
                or occurs('MATLAB:nomem.', task_stderr, fgrep))
            if generic_memory_error or matlab_memory_error:
                new_requested_memory = requested_memory + self.increment
                if new_requested_memory >= self.maximum:
                    gc3libs.log.info(
                        "%s: Possible out-of-memory condition detected,"
                        " but increasing memory requirements would"
                        " exceed set maximum of %s.  Aborting task.",
                        self.task, self.maximum)
                    return False
                else:
                    self.task.requested_memory = new_requested_memory
                    gc3libs.log.info(
                        "%s: Possible out-of-memory condition detected,"
                        " will request %s for next run.",
                        self.task, self.task.requested_memory)
                    return True
            else:
                gc3libs.log.info(
                    "%s: Task failed for non-memory-related reasons,"
                    " *not* resubmitting it again.", self.task)
                return False


class FunctionalDiversityApplication(Application):
    """
    Custom class to wrap the execution of the "functional diversity"
    MATLAB code by Fabian Daniel Schneider.
    """

    application_name = 'functional_diversity'

    # pattern for the MATLAB commands to run
    matlab_cmd = (
        "load('{inputname}.mat');"
        "outputData={funcname}({params},{radius});"
        "save('{outputfile}','outputData');"
    )

    def __init__(self, funcfile, radius, params,
                 inputfile, outputfile=None, **extra_args):
        funcname = basename_sans(funcfile)
        self.funcname = funcname
        self.radius = radius
        # map args to file system names
        inputname = basename_sans(inputfile)
        if outputfile is None:
            outputfile = ('output_{inputname}_{radius}.mat'.format(**locals()))
        # default execution params
        extra_args.setdefault('requested_cores',        1)
        extra_args.setdefault('requested_memory',       3*GB)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args.setdefault('requested_walltime',     30*days)
        # actual app initialization
        Application.__init__(
            self,
            arguments=[
                'matlab', '-nodisplay', '-nojvm', #'-singleCompThread',
                '-r', (self.matlab_cmd.format(**locals()))
            ],
            inputs = [funcfile, inputfile],
            outputs = [outputfile],
            stdout = 'matlab.log',
            join=True,
            **extra_args)


## main script class

class GfdivScript(SessionBasedScript):
    """
Run the specified MATLAB function on all the given input files,
for all neighborhood radius parameter in the given range.

The ``gfdiv`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files or an extended range is specified in the command line.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = FunctionalDiversityApplication,
        )


    def setup_options(self):
        self.add_param(
            '-p', '--parameters', default='inputData',
            help=(
                "List of argument names to pass to the MATLAB function."
                " The neighborhood radius is always appended at the end."))
        # change default for the memory/walltime options
        self.actions['memory_per_core'].default = 3*Memory.GB
        self.actions['wctime'].default = '60 days'


    def setup_args(self):
        self.add_param(
            'args', nargs='+',
            help=(
                "A series of arguments for building a composite session:"
                " consists of one or more triplets of parameters, separated"
                " by a literal `::`."
                ""
                " Each triplet consists of the following items:"
                ""
                " - Name of the MATLAB function to apply to the input data."
                "   A corresponding `.m` file must exist in the current directory."
                ""
                " - Values for the neighborhood radius parameter."
                "   Can be any of the following:"
                "   (1) a single numeric value;"
                "   (2) a range MIN:MAX (e.g., 1:170) to compute all radii"
                "       from MIN to MAX (inclusive).  Optionally,"
                "       the MIN:MAX:STEP form can be used to further specify"
                "       an increment; e.g., 1:170:10 would run computations"
                "       with radius 1,11,21,...,161;"
                "   (3) a comma-separated list of the above two forms."
                ""
                " - Input data file(s)."
            ))


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    @staticmethod
    def get_function_name_and_file(funcname):
        if funcname.endswith('.m'):
            funcname_m = funcname
            funcname = funcname[:-len('.m')]
        else:
            funcname_m = funcname + '.m'
        funcfile = join(os.getcwd(), funcname_m)
        if not exists(funcfile):
            raise ArgumentError(
                self.actions['funcname'],
                ("Cannot read file '{funcfile}'"
                 " providing MATLAB function '{funcname}'.")
                .format(**locals()))
        return funcname, funcfile


    def new_tasks(self, extra):
        while self.params.args:
            # get next chunk of arguments
            if '::' in self.params.args:
                up_to = self.params.args.index('::')
                args = self.params.args[:up_to]
                self.params.args = self.params.args[(up_to+1):]
            else:
                args = self.params.args
                self.params.args = []
            # process it
            assert len(args) > 2
            funcname = args[0]
            values = args[1]
            inputfiles = args[2:]
            for inputfile in inputfiles:
                assert existing_file(inputfile)
            # add tasks to session
            for task in self.new_tasks1(funcname, values, inputfiles, **extra):
                yield task


    def new_tasks1(self, funcname, values, inputfiles, **extra):
        """
        Iterate over tasks with a common processing function.
        """
        # find MATLAB function to run
        funcname, funcfile = self.get_function_name_and_file(funcname)
        ranges = []
        for spec in values.split(','):
            # get range for neighborhood radius
            low, high, step = parse_range(spec)
            ranges.append(irange(low, high+1, step))
        for radius in itertools.chain(*ranges):
            for inputfile in inputfiles:
                kwargs = extra.copy()
                base_output_dir = kwargs.pop('output_dir', self.params.output)
                inputfile = realpath(inputfile)
                inputname = basename_sans(inputfile)
                jobname = ('{funcname}_{inputname}_{radius}'.format(**locals()))
                outputfile = ('output_{inputname}_{radius}.mat'.format(**locals()))
                output_dir = join(base_output_dir, jobname)
                yield MatlabRetryOnOutOfMemory(
                    FunctionalDiversityApplication(
                        funcfile,
                        radius,
                        self.params.parameters,
                        inputfile,
                        jobname=jobname,
                        outputfile=outputfile,
                        output_dir=output_dir,
                        **kwargs),
                    increment=4*GB,
                    jobname=jobname,
                )



#! /usr/bin/env python
#
#   gsubbeast.py -- Front-end script for running BEAST jobs.
#   https://www.beast2.org/
#
#   Copyright (C) 2017 2018  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple BEAST jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsubbeast.py --help`` for program usage
instructions.

Input parameters consists of:
:param str BEAST XML file: Path to an .xml file containing BEAST settings
and input data
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2017-08-28:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gsubbeast
    gsubbeast.GsubbeastScript().run()

import os
import sys
import time
import tempfile
import mimetypes
import random

import pandas
from xml.etree import cElementTree as ET
from xml.etree.ElementTree import Element, SubElement, Comment, tostring

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_REMOTE_OUTPUT_FOLDER = "./results"
BEAST_COMMAND="sudo docker run -v $PWD:/data smaffiol/beast2:2.4.7 -working -overwrite -beagle {resume} -threads -1 -instances -1 /data/{input_xml}"

# Utility functions
def _get_id_from_inputfile(input_file):
    """
    by convention, input XML files will have an 'id' tag in the 'data' item
    Return 'id' tag name
    """
    try:
        id_name = filename = os.path.splitext(os.path.basename(input_file))[0]
        tree = ET.parse(input_file)
        root = tree.getroot()
        data = root.find('data')
        if data.get('id'):
            id_name = data.get('id')
        else:
            gc3libs.log.warning("Failed to extract 'id' from input file {0}. Using filename instead".format(filename))
            gc3libs.log.info("Using id '{0}' for filename {1}".format(id_name, filename))
        return id_name
    except ET.ParseError as px:
        gc3libs.log.error("Failed parsing input file {0}. Error {1}".format(input_file,
                                                                            px.message))
        return None

def _get_valid_input(input_folder, resume):
    """
    Returns tuple (string,int,string):  (filename,id_name,resume_file)
    each returned tuple should reference a valid input .xml filename
    """

    state_file = None
    for input_file in [ os.path.join(input_folder,input_xml) for input_xml in os.listdir(input_folder) if mimetypes.guess_type(input_xml)[0] == 'application/xml' ]:
        # extract ID name. If fails, skip file
        id_name = _get_id_from_inputfile(input_file)
        if not id_name:
            gc3libs.log.warning("skipping {0}".format(input_file))
            continue
        # check whether statefile is present. If so, include in the returned tuple
        if resume and os.path.isfile(os.path.join(input_folder,input_file+'.state')):
            state_file = (input_file+'.state')

        # Return tuple (filename,id_name,resume_file)
        yield (input_file,id_name,state_file)


def _check_exit_condition(log, output_dir):
    """
    Inspect output folder.
    Check for termiantion condition AND .state file
    If termination condition not met, return list of .state files
    If termiantion condition is met, return an empty list.
    """
    # XXX: weak! this is very dependent on beast stderr formats
    TERMIANTION_PATTERN = "End likelihood"
    with open(log) as fd:
        for line in fd:
            if TERMIANTION_PATTERN in line:
                # Job completed.
                # Return an empty list
                return (None,None)

    # Somehow job was not completed.
    # Search for .state file and return it
    results = os.listdir(output_dir)
    for item in results:
        if item.endswith(".state"):
            return (os.path.join(output_dir,item),results)

    return (None,None)

## custom application class
class GsubbeastApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gsubbeast'

    def __init__(self, input_file, state_file, id_name, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        inputs[input_file] = os.path.basename(input_file)

        if state_file:
            inputs[state_file] = os.path.basename(state_file)
            resume_option=" -t {state_file}".format(state_file=state_file)
        else:
            resume_option=""

        arguments = BEAST_COMMAND.format(resume=resume_option,
                                         input_xml=inputs[input_file])

        gc3libs.log.debug("Creating application for executing: %s", arguments)

        self.id_name = id_name

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gsubbeast.log',
            join=True,
            executables = executables,
            **extra_args)


    def terminated(self):
        """
        Check whehter expected output log files have been generated.
        """
        result_log_file = os.path.join(self.output_dir, '{0}.log'.format(self.id_name))
        gc3libs.log.info('Application terminated with exit code %s' % self.execution.exitcode)
        if not os.path.isfile(result_log_file):
            gc3libs.log.error('Failed while checking outputfile %s.' % result_log_file)
            self.execution.returncode = (0, 99)

class GsubbeastScript(SessionBasedScript):
    """
    The ``gsubbeast`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsubbeast`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gsubbeast``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GsubbeastApplication,
            stats_only_for = GsubbeastApplication,
            )

    def setup_args(self):
        self.add_param('input_folder',
                       type=existing_directory,
                       help="Path to input folder containing valid input .xml files.")

    def setup_options(self):

        self.add_param("-U", "--resume",
                       action="store_true",
                       dest="resume",
                       default=False,
                       help="Use existing '.state' files to "
                       "resume interrupted BEAST execution. "
                       "Default: %(default)s.")

        self.add_param("-D", "--docker", metavar="[STRING]",
                       dest="dockerimage",
                       default=None,
                       help="Docker image to use.")

        self.add_param("-F", "--follow",
                       dest="follow",
                       action="store_true",
                       default=False,
                       help="Periodically fetch job's output folder and copy locally.")

        self.add_param('-M', '--merge-anyway',
                       dest='merge_anyway',
                       action='store_true',
                       default=False,
                       help="Merge results only when all jobs have completed " \
                       " successfully. Default: %(default)s.")

        self.add_param('-O', '--store-aggregate-csv',
                       dest='result_csv',
                       default='.',
                       help="Location of aggregated .csv results. Default: '%(default)s'.")

        self.add_param('-P', '--extract-columns',
                       dest='columns',
                       default='TreeHeight.t',
                       help='Comma separated list of columns name to extract from " \
                       "output log file. Default: %(default)s.')

    def parse_args(self):
        self.params.columns = self.params.columns.split(',')

        if not os.path.isdir(self.params.result_csv):
            gc3libs.log.info("Creating CSV result folder: '{0}'".format(self.params.result_csv))
            os.makedirs(self.params.result_csv)


    def before_main_loop(self):
        # XXX: should this be done with `make_controller` instead?
        self._controller.retrieve_running = self.params.follow
        self._controller.retrieve_overwrites = self.params.follow
        self._controller.retrieve_changed_only = self.params.follow

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """
        tasks = []

        for (input_file, id_name, stat_file) in _get_valid_input(self.params.input_folder,
                                                                 self.params.resume):
            extra_args = extra.copy()
            extra_args['jobname'] = id_name
            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', extra_args['jobname'])
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', extra_args['jobname'])

            self.log.debug("Creating Application for file '%s'" % extra_args['jobname'])

            tasks.append(GsubbeastApplication(
                input_file,
                stat_file,
                id_name,
                **extra_args))

        return tasks

    def after_main_loop(self):
        """
        Merge all result files together
        Then clean up tmp files
        """
        if not self.params.merge_anyway:
            for task in self.session:
                if task.execution.state != Run.State.TERMINATED:
                    gc3libs.log.warning('Could not perform aggregation task as not all jobs have terminated.')
                    return
                if task.execution.returncode is not None and task.execution.returncode != 0:
                    gc3libs.log.warning('Could not perform aggregation task as not all jobs have completed successfully.')
                    return

        df_dict = dict()
        for df_name in self.params.columns:
            df_dict[df_name] = pandas.DataFrame()

        for task in self.session:
            if isinstance(task, GsubbeastApplication) and task.execution.returncode == 0:
                result_log_file = os.path.join(task.output_dir, '{0}.log'.format(task.id_name))
                if os.path.isfile(result_log_file):
                    gc3libs.log.debug('Reading output file {0}'.format(result_log_file))
                    data = pandas.read_csv(result_log_file, sep='\t', comment="#")
                    for key in df_dict.keys():
                        cols = [col for col in data.columns if key in col]
                        # In case multiple entries, take the first occurrence
                        # column_to_search = "{0}:{1}".format(key,task.id_name)
                        gc3libs.log.debug("Column [{0}] found {1} occurances. Should be 1.".format(key,len(cols)))
                        if len(cols) > 0:
                            df_dict[key][task.id_name] = data[cols[0]]
                        else:
                            gc3libs.log.error("Skipping column {0} as no occourrences have been found".format(key))

                else:
                    gc3libs.log.error('Output file {0} for task {1} not found'.format(result_log_file, task.id_name))
                    continue
            else:
                gc3libs.log.warning('Task {0} not completed or failed during execution. Ignoring'.format(task.id_name))

        gc3libs.log.info('Writing aggregated .csv results in {0}'.format(self.params.result_csv))
        for key in df_dict.keys():
            df = df_dict[key]
            df.to_csv(os.path.join(self.params.result_csv, '{0}.csv'.format(key)))

        return



#! /usr/bin/env python
#
#   gstructure.py -- Front-end script for submitting multiple `gstructure` jobs.
#
#   Copyright (C) 2013, 2014  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `Structure` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gstructure --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """

  2014-02-27:
  * Initial release, forked off the ``gpyrad`` sources.
"""
__author__ = 'Tyanko Aleksiev <tyanko.aleksiev@chem.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gstructure
    gstructure.GStructureScript().run()

from pkg_resources import Requirement, resource_filename
import os
import csv
import posix

# gc3 library imports
import gc3libs.utils
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class

class GStructureApplication(Application):
    """
    """

    application_name = 'structure'

    def __init__(self, input_file, **extra_args):
        """
        The wrapper script is being used for start the simulation.
        """
        files_to_send = []

        gstructure_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/gstructure_wrapper.sh")

        basename_input_file = os.path.basename(input_file)
        files_to_send.append((gstructure_wrapper_sh,os.path.basename(gstructure_wrapper_sh)))
        files_to_send.append((input_file,basename_input_file))


        cmd = "./gstructure_wrapper.sh -d "

        if 'mainparam_file' in extra_args:
            cmd += " -p %s " % extra_args['mainparam_file']
            files_to_send.append(extra_args['mainparam_file'],'mainparams.txt')

        if 'extraparam_file' in extra_args:
            cmd += " -x %s " % extra_args['extraparam_file']
            files_to_send.append(extra_args['extraparam_file'],'extraparams.txt')

        if 'output_file' in extra_args:
            cmd += " -u %s " % extra_args['output_file']
        else:
            output_file = basename_input_file.split(".")[0] + ".out"
            cmd += " -u %s " % output_file

        if extra_args.has_key('k_range'):
            cmd += " -g %s " % extra_args['k_range']
        else:
            cmd += " -g %s " % extra_args['k_range'].default

        if extra_args.has_key('replica'):
            cmd += " -e %s " % extra_args['replica']
        else:
            cmd += " -e %s " % extra_args['replica'].default

        cmd += " %s " % extra_args['loc']

        cmd += " %s " % extra_args['ind']

        cmd += " %s " % basename_input_file

        extra_args['requested_memory'] = 7*GB

        self.output_dir = basename_input_file + "_output"
        extra_args['output_dir'] = self.output_dir

        Application.__init__(
            self,
            # arguments should mimic the command line interfaca of the command to be
            # executed on the remote end
            arguments = cmd,
            inputs = files_to_send,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gstructure.log',
            join=True,
            **extra_args)

class GStructureTask(RetryableTask, gc3libs.utils.Struct):
    """
    Run ``gstructure`` on a given simulation directory until completion.
    """
    def __init__(self, input_file, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GStructureApplication(input_file, **extra_args),
            # keyword arguments
            **extra_args)

## main script class

class GStructureScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``gstructure`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gstructure`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GStructureTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GStructureTask,
            )


    def setup_options(self):

        self.add_param("-p", "--mainparams", metavar="MAIN_CONFIG_FILE", default="mainparams.txt",
                       dest="mainparams_file", help="Uses a different main parameters file.")

        self.add_param("-x", "--extraparams", metavar="EXTRA_CONFIG_FILE", default="extraparams.txt",
                       dest="extraparams_file", help="Uses a different extra parameters file.")

        self.add_param("-u", "--output", metavar="OUTPUT_FILE_NAME",
                       dest="output_file", help="Output file name where results will be saved. "
                       "Default: INPUT_FILE.out" )

        self.add_param("-g", "--K-range", metavar="K_RANGE", default="1:20",
                       dest="k_range", help="Structure K range. "
                       "Default: %(default)s" )

        self.add_param("-e", "--replica", metavar="REPLICA_NUM", default=3,
                       dest="replica", help="Structure replicates. "
                       "Default: %(default)s" )

        self.add_param("-T", "--control-file", metavar="CONTROL_FILE",
                       dest="control_file", help="Control csv file for managing multiple input files with differetn loc and ind")

        self.add_param("--loc", type=int, metavar="LOC",
                       dest="loc", help="Number of loci in the structure input file")

        self.add_param("--ind", type=int, metavar="IND",
                       dest="ind", help="Number of individuals in the structure input file")

        self.add_param("--input_source", type=str,metavar="INPUT_SOURCE",
                       dest="input_source", help="Structure input file/Structure input directory")

    def parse_args(self):

        if self.params.input_source:
            self.params.input_source = os.path.abspath(self.params.input_source)

    def new_tasks(self, extra):

        tasks = []
        input_files = []
        inputs = []
        extentions = ( '.tsv', '.txt', '.struc', '.tab' )
        extra_args = extra.copy()

        if self.params.mainparams_file:
            extra_args['mainparams_file'] = self.params.mainparams_file

        if self.params.extraparams_file:
            extra_args['extraparams_file'] = self.params.extraparams_file

        if self.params.k_range:
            extra_args['k_range'] = self.params.k_range

        if self.params.replica:
            extra_args['replica'] = self.params.replica

        # Check if control_file variable is defined. If yes,
        # proceed differently for crating the tasks.
        if self.params.control_file:
            gc3libs.log.warning("As you issued the script with the control, file options loc and ind paramters will be ignored")

            if self.params.control_file.endswith('.csv'):
                try:
                    inputfile = open(self.params.control_file, 'r')
                except (OSError, IOError), ex:
                    self.log.warning("Cannot open input file '%s': %s: %s",
                                     path, ex.__class__.__name__, str(ex))
                for row in csv.reader(inputfile):
                    extra_args['input_source'] = row[0]
                    extra_args['loc'] = row[1]
                    extra_args['ind'] = row[2]

                    input_file = row[0]

                    tasks.append(GStructureTask(
                        input_file,
                        **extra_args
                        ))
        else:
             if os.path.isdir(self.params.input_source):

                for i in self._list_local_folder(self.params.input_source):
                    for ext in extentions:
                        if i.endswith(ext):
                            input_files.append(i)

             elif os.path.isfile(self.params.input_source):

                for ext in extentions:
                        if self.params.input_source.endswith(ext):
                            input_files.append(self.params.input_source)

             for input_file in input_files:

                    jobname = "%s" % input_file.split(".")[0]

                    extra_args['jobname'] = jobname

                    # FIXME: ignore SessionBasedScript feature of customizing
                    # output folder

                    extra_args['loc'] = self.params.loc
                    extra_args['ind'] = self.params.ind
                    extra_args['input_source'] = self.params.input_source
                    #extra_args['output_dir'] = self.params.input_source

                    self.log.info("Creating Task for input file: %s" % input_file)

                    tasks.append(GStructureTask(
                        input_file,
                        **extra_args
                        ))

        return tasks


    def _list_local_folder(self, input_folder):
        """
        return a list of all files in the input folder
        """

        return [ os.path.join(input_folder,infile) for infile in os.listdir(input_folder) ]



#! /usr/bin/env python
#
"""
"""
from __future__ import division
# Copyright (C) 2012-2013,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
from future import standard_library
standard_library.install_aliases()
from past.utils import old_div
__docformat__ = 'reStructuredText'

from __future__ import absolute_import, print_function, unicode_literals
import argparse
import json
import logging
import mimetypes
import os
import sys
import time
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse

try:
    import magic
except ImportError:
    magic = None

log = logging.getLogger()
log.addHandler(logging.StreamHandler(sys.stderr))
log.setLevel(logging.DEBUG)

def open_http(url):
    log.info("Sending request to %s", url.geturl())
    return urllib.request.urlopen(url.geturl())


def open_swift(url, method='get', content_type=None, content_length=0, content_data=None):
    username, tenant = url.username.split('+')
    password = url.password
    if url.scheme in ['swifts', 'swts']:
        auth_url = 'https://%s' % url.hostname
    else:
        auth_url = 'http://%s' % url.hostname
    query = url.query.split('&')
    container = query[0]
    object_name = query[1].rsplit('=',1)[-1]
    if url.port:
        auth_url += ':%d' % url.port
    if url.path:
        auth_url += url.path

    # Get a token from keystone
    data = json.dumps(
        {
            'auth' : {
                'tenantName' : tenant,
                'passwordCredentials' : {
                    'username' : username,
                    'password' : password
                }
            }
        }
    )
    token_url = auth_url+'/tokens'
    log.info("Getting token from '%s' for user '%s', tenant '%s'",
             token_url, username, tenant)

    kreq = urllib.request.Request(token_url, data)
    kreq.add_header("Content-type", "application/json")

    fp = urllib.request.urlopen(kreq)
    authresp = json.loads(fp.read())
    fp.close()

    token = authresp['access']['token']['id']

    # Get the storage URL
    for endpoint in authresp['access']['serviceCatalog']:
        if endpoint['type'] == 'object-store':
            storage_url = endpoint['endpoints'][0]['publicURL']
    log.info("Token recovered: %s, storage_url: %s",
              token, storage_url)
    # Get the object from swift
    object_url = os.path.join(
        storage_url,
        container,
        object_name).encode('utf-8')

    # If this is a GET, we just download it
    sreq = urllib.request.Request(object_url, content_data)
    sreq.add_header('X-Auth-Token', token)

    # If it's a PUT, we also need to set more information
    if method.upper() == 'PUT':
        sreq.get_method = lambda: 'PUT'
        sreq.add_header('Content-Length', len(content_data))
        if content_type:
            sreq.add_header('Content-Type', content_type)
        else:
            sreq.add_header('Content-Type', 'application/octect-stream')
    log.info("Headers: %s" % sreq.headers)
    log.info("Sending request to %s", object_url)
    return urllib.request.urlopen(sreq)


def download_file(url, outfile, bufsize=2**20):
    url = urllib2.urlparse.urlparse(url)
    stime = time.time()
    if url.scheme in ['http', 'https']:
        fd = open_http(url)
    elif url.scheme in ['swift', 'swt', 'swifts', 'swts']:
        fd = open_swift(url)
    if not fd:
        log.error("Unrecognized scheme %s", url.scheme)
        return 1

    with open(outfile, 'w') as outfd:
        data = fd.read(bufsize)
        while data:
            outfd.write(data)
            data = fd.read(bufsize)
    fd.close()
    etime = time.time()
    log.info("File '%s' downloaded to '%s' in %f seconds (%d bytes/s)", url.geturl(), outfile, etime-stime, old_div(os.stat(outfile).st_size,(etime-stime)))

def upload_file(url, local, bufsize=2**20):
    url = urllib2.urlparse.urlparse(url)
    stime = time.time()
    if url.scheme in ['http', 'https']:
        log.error("Unable to upload to '%s'", url.scheme)
        return 1

    if url.scheme in ['swift', 'swifts', 'swt', 'swts']:
        # Guess the content-type of the file.
        try:
            # magic module might not be installed, but it works way
            # better than `mimetypes`
            m = magic.open(magic.MAGIC_MIME)
            m.load()
            ctype = m.file(local)
        except:
            # revert back to mime types, which is only checking the
            # file extension
            ctype = mimetypes.guess_type(local)[0]
        # length can be
        clength = os.stat(local).st_size
        with open(local, 'r') as localfd:
            fd = open_swift(url,
                            method='put',
                            content_type=ctype,
                            content_length=clength,
                            content_data=localfd.read())
    etime = time.time()
    log.info("File '%s' uploaded to '%s' in %f seconds (%d bytes/s)", local, url.geturl(), etime-stime, old_div(clength,(etime-stime)))


## main: run tests

if "__main__" == __name__:
    parser = argparse.ArgumentParser()
    parser.add_argument('action', choices=['upload', 'download'])
    parser.add_argument('remote')
    parser.add_argument('local')
    cfg = parser.parse_args()
    if cfg.action == 'download':
        sys.exit(download_file(cfg.remote, cfg.local))
    else:
        if not os.path.isfile(cfg.local):
            parser.error("File '%s' not found. Unable to upload it." % cfg.local)
        sys.exit(upload_file(cfg.remote, cfg.local))
